# -*- coding: utf-8 -*-
import onegov.core
import onegov.town

from mock import patch
from morepath import setup
from onegov.testing import utils
from webtest import TestApp as Client
from webtest import Upload


@patch('morepath.directive.register_view')
def test_view_permissions(register_view):
    config = setup()
    config.scan(onegov.core)
    config.scan(onegov.town)
    config.commit()

    # make sure that all registered views have an explicit permission
    for call in register_view.call_args_list:
        view = call[0][2]
        module = view.__venusian_callbacks__[None][0][1]
        permission = call[0][5]

        if module.startswith('onegov.town') and permission is None:
            assert permission is not None, (
                'view {}.{} has no permission'.format(module, view.__name__))


def test_view_login(town_app):

    client = Client(town_app)

    assert client.get('/logout', expect_errors=True).status_code == 403

    response = client.get('/login')

    # German is the default translation and there's no English translation yet
    # (the default *is* English, but it needs to be added as a locale, or it
    # won't be used)
    assert response.status_code == 200
    assert u"E-Mail Adresse" in response
    assert u"Passwort" in response

    assert 'userid' not in client.cookies
    assert 'role' not in client.cookies
    assert 'application_id' not in client.cookies
    assert client.get('/logout', expect_errors=True).status_code == 403

    response.form.set('email', 'admin@example.org')
    response = response.form.submit()
    assert response.status_code == 200
    assert u"E-Mail Adresse" in response
    assert u"Passwort" in response
    assert u"Dieses Feld wird benötigt." in response

    assert 'userid' not in client.cookies
    assert 'role' not in client.cookies
    assert 'application_id' not in client.cookies
    assert client.get('/logout', expect_errors=True).status_code == 403

    response.form.set('email', 'admin@example.org')
    response.form.set('password', 'hunter2')
    response = response.form.submit()
    assert response.status_code == 302

    assert 'userid' in client.cookies
    assert 'role' in client.cookies
    assert 'application_id' in client.cookies

    assert client.get('/logout').status_code == 302
    assert 'userid' not in client.cookies
    assert 'role' not in client.cookies
    assert 'application_id' not in client.cookies

    assert client.get('/logout', expect_errors=True).status_code == 403


def test_view_images(town_app):

    client = Client(town_app)

    assert client.get('/bilder', expect_errors=True).status_code == 403

    login_page = client.get('/login')
    login_page.form.set('email', 'admin@example.org')
    login_page.form.set('password', 'hunter2')
    login_page.form.submit()

    images_page = client.get('/bilder')

    assert "Noch keine Bilder hochgeladen" in images_page

    images_page.form['file'] = Upload('Test.txt', utils.create_image().read())
    assert images_page.form.submit(expect_errors=True).status_code == 415

    images_page.form['file'] = Upload('Test.jpg', utils.create_image().read())
    images_page = images_page.form.submit().follow()

    assert "Noch keine Bilder hochgeladen" not in images_page

    name = images_page.pyquery('img').attr('src').split('/')[-1]

    # thumbnails are created on the fly
    assert town_app.filestorage.exists('images/' + name)
    assert not town_app.filestorage.exists('images/thumbnails' + name)
    client.get('/thumbnails/' + name)
    assert town_app.filestorage.exists('images/thumbnails/' + name)

    # thumbnails are deleted with the image
    delete_link = images_page.pyquery('a.delete').attr('ic-delete-from')
    client.delete(delete_link)
    assert not town_app.filestorage.exists('images/' + name)
    assert not town_app.filestorage.exists('images/thumbnails/' + name)


def test_startpage(town_app):
    client = Client(town_app)

    links = client.get('/').pyquery('.top-bar-section a')

    links[0].text == 'Leben & Wohnen'
    links[0].attrib.get('href') == '/gemeinde/leben-wohnen'

    links[1].text == 'Kultur & Freizeit'
    links[1].attrib.get('href') == '/gemeinde/kultur-freizeit'

    links[2].text == 'Bildung & Gesellschaft'
    links[2].attrib.get('href') == '/gemeinde/bildung-gesellschaft'

    links[3].text == 'Gewerbe & Tourismus'
    links[3].attrib.get('href') == '/gemeinde/gewerbe-tourismus'

    links[4].text == 'Politik & Verwaltung'
    links[4].attrib.get('href') == '/gemeinde/politik-verwaltung'

    links = client.get('/').pyquery('.homepage-tiles a')

    links[0].text == 'Leben & Wohnen'
    links[0].attrib.get('href') == '/gemeinde/leben-wohnen'

    links[1].text == 'Kultur & Freizeit'
    links[1].attrib.get('href') == '/gemeinde/kultur-freizeit'

    links[2].text == 'Bildung & Gesellschaft'
    links[2].attrib.get('href') == '/gemeinde/bildung-gesellschaft'

    links[3].text == 'Gewerbe & Tourismus'
    links[3].attrib.get('href') == '/gemeinde/gewerbe-tourismus'

    links[4].text == 'Politik & Verwaltung'
    links[4].attrib.get('href') == '/gemeinde/politik-verwaltung'


def test_login(town_app):
    client = Client(town_app)

    links = client.get('/').pyquery('.bottom-links a:first-child')
    assert links.text() == 'Login'

    login_page = client.get(links.attr('href'))
    login_page.form['email'] = 'admin@example.org'
    login_page.form['password'] = ''
    login_page = login_page.form.submit()

    assert u"Dieses Feld wird benötigt" in login_page.text

    login_page.form['email'] = 'admin@example.org'
    login_page.form['password'] = 'wrong'
    login_page = login_page.form.submit()

    assert "Unbekannter Benutzername oder falsches Passwort" in login_page.text

    login_page.form['email'] = 'admin@example.org'
    login_page.form['password'] = 'hunter2'

    index_page = login_page.form.submit().follow()
    assert "Sie wurden eingeloggt" in index_page.text

    links = index_page.pyquery('.bottom-links a:first-child')
    assert links.text() == 'Logout'

    index_page = client.get(links.attr('href')).follow()
    assert "Sie wurden ausgeloggt" in index_page.text

    links = index_page.pyquery('.bottom-links a:first-child')
    assert links.text() == 'Login'


def test_settings(town_app):
    client = Client(town_app)

    assert client.get('/einstellungen', expect_errors=True).status_code == 403

    login_page = client.get('/login')
    login_page.form['email'] = 'admin@example.org'
    login_page.form['password'] = 'hunter2'
    login_page.form.submit()

    settings_page = client.get('/einstellungen')
    document = settings_page.pyquery

    assert document.find('input[name=name]').val() == 'Govikon'
    assert document.find('input[name=primary_color]').val() == '#006fba'

    settings_page.form['primary_color'] = '#xxx'
    settings_page = settings_page.form.submit()

    assert u"Ungültige Farbe." in settings_page.text

    settings_page.form['primary_color'] = '#ccddee'
    settings_page = settings_page.form.submit()

    assert u"Ungültige Farbe." not in settings_page.text

    settings_page.form['logo_url'] = 'https://seantis.ch/logo.img'
    settings_page = settings_page.form.submit()

    assert '<img src="https://seantis.ch/logo.img"' in settings_page.text

    settings_page.form['homepage_images'] = """
        http://images/one
        http://images/two
    """
    settings_page = settings_page.form.submit()

    assert 'http://images/one' in settings_page
    assert 'http://images/two' in settings_page


def test_unauthorized(town_app):
    client = Client(town_app)

    unauth_page = client.get('/einstellungen', expect_errors=True)
    assert u"Zugriff verweigert" in unauth_page.text
    assert u"folgen Sie diesem Link um sich anzumelden" in unauth_page.text

    link = unauth_page.pyquery('#alternate-login-link')[0]
    login_page = client.get(link.attrib.get('href'))
    login_page.form['email'] = 'editor@example.org'
    login_page.form['password'] = 'hunter2'
    unauth_page = login_page.form.submit().follow(expect_errors=True)

    assert u"Zugriff verweigert" in unauth_page.text
    assert u"mit einem anderen Benutzer anzumelden" in unauth_page.text

    link = unauth_page.pyquery('#alternate-login-link')[0]
    login_page = client.get(link.attrib.get('href'))
    login_page.form['email'] = 'admin@example.org'
    login_page.form['password'] = 'hunter2'
    settings_page = login_page.form.submit().follow()

    assert settings_page.status_code == 200
    assert u"Zugriff verweigert" not in settings_page


def test_pages(town_app):
    client = Client(town_app)

    root_url = client.get('/').pyquery('.top-bar-section a').attr('href')
    assert len(client.get(root_url).pyquery('.edit-bar')) == 0

    login_page = client.get('/login?to={}'.format(root_url))
    login_page.form['email'] = 'admin@example.org'
    login_page.form['password'] = 'hunter2'
    root_page = login_page.form.submit().follow()

    assert len(root_page.pyquery('.edit-bar')) == 1
    new_page = root_page.click('Thema')
    assert "Neues Thema" in new_page

    new_page.form['title'] = "Living in Govikon is Swell"
    new_page.form['text'] = (
        "<h2>Living in Govikon is Really Great</h2>"
        "<i>Experts say it's the fact that Govikon does not really exist.</i>"
    )
    page = new_page.form.submit().follow()

    assert page.pyquery('.main-title').text() == "Living in Govikon is Swell"
    assert page.pyquery('h2').text() == "Living in Govikon is Really Great"
    assert page.pyquery('i').text().startswith("Experts say it's the fact")

    edit_page = page.click("Bearbeiten")

    assert "Thema Bearbeiten" in edit_page
    assert "&lt;h2&gt;Living in Govikon is Really Great&lt;/h2&gt" in edit_page

    edit_page.form['title'] = "Living in Govikon is Awful"
    edit_page.form['text'] = (
        "<h2>Living in Govikon Really Sucks</h2>"
        "<i>Experts say hiring more experts would help.</i>"
        "<script>alert('yes')</script>"
    )
    page = edit_page.form.submit().follow()

    assert page.pyquery('.main-title').text() == "Living in Govikon is Awful"
    assert page.pyquery('h2').text() == "Living in Govikon Really Sucks"
    assert page.pyquery('i').text().startswith("Experts say hiring more")
    assert "<script>alert('yes')</script>" not in page
    assert "&lt;script&gt;alert('yes')&lt;/script&gt;" in page

    page.click("Logout")
    root_page = client.get(root_url)

    assert len(root_page.pyquery('.edit-bar')) == 0

    assert page.pyquery('.main-title').text() == "Living in Govikon is Awful"
    assert page.pyquery('h2').text() == "Living in Govikon Really Sucks"
    assert page.pyquery('i').text().startswith("Experts say hiring more")


def test_delete_pages(town_app):
    client = Client(town_app)

    root_url = client.get('/').pyquery('.top-bar-section a').attr('href')

    login_page = client.get('/login?to={}'.format(root_url))
    login_page.form['email'] = 'admin@example.org'
    login_page.form['password'] = 'hunter2'

    root_page = login_page.form.submit().follow()
    new_page = root_page.click('Thema')

    new_page.form['title'] = "Living in Govikon is Swell"
    new_page.form['text'] = (
        "## Living in Govikon is Really Great\n"
        "*Experts say it's the fact that Govikon does not really exist.*"
    )
    page = new_page.form.submit().follow()
    delete_link = page.pyquery('a[ic-delete-from]')[0].attrib['ic-delete-from']

    result = client.delete(delete_link.split('?')[0], expect_errors=True)
    assert result.status_code == 403

    assert client.delete(delete_link).status_code == 200
    assert client.delete(delete_link, expect_errors=True).status_code == 404


def test_links(town_app):
    client = Client(town_app)

    root_url = client.get('/').pyquery('.top-bar-section a').attr('href')

    login_page = client.get('/login?to={}'.format(root_url))
    login_page.form['email'] = 'admin@example.org'
    login_page.form['password'] = 'hunter2'
    root_page = login_page.form.submit().follow()

    new_link = root_page.click(u"Verknüpfung")
    assert u"Neue Verknüpfung" in new_link

    new_link.form['title'] = 'Google'
    new_link.form['url'] = 'https://www.google.ch'
    link = new_link.form.submit().follow()

    assert "Sie wurden nicht automatisch weitergeleitet" in link
    assert 'https://www.google.ch' in link

    link.click('Logout')

    root_page = client.get(root_url)
    assert "Google" in root_page
    google = root_page.click("Google", index=0)

    assert google.status_code == 302
    assert google.location == 'https://www.google.ch'



from onegov.foundation import BaseTheme
from onegov.core.utils import module_path


# the colors which are editable by the user
user_colors = {
    'primary-color': '#006fba'
}


class TownTheme(BaseTheme):
    name = 'onegov.town.foundation'

    # don't touch this number, it's incremented using bumpversion, so every
    # release will automatically trigger a rebuild of the theme
    version = '0.0.2'

    @property
    def default_options(self):
        options = {
            'top-bar-border-size': '0.3rem',
            'bottom-links-color': '#777',
            'bottom-links-size': '0.8rem',
            'topbar-bg-color': '#f5f5f5',
            'topbar-link-bg-hover': '#dfdfdf',
            'topbar-link-color': '#312f2e',
            'topbar-link-color-hover': '#312f2e',
            'topbar-link-color-active': '#312f2e',
            'topbar-link-color-active-hover': '#312f2e',
            'topbar-link-weight': 'bold',
            'topbar-menu-link-color': '#312f2e',
            'topbar-menu-icon-color': '#312f2e',
            'topbar-dropdown-bg': '#f5f5f5',
            'side-nav-font-weight-active': 'bold',
            'crumb-bg': '#fff',
            'crumb-border-size': '0',
            'tile-image-1': '"../static/homepage-images/tile-1-small.jpg"',
            'tile-image-2': '"../static/homepage-images/tile-2-small.jpg"',
            'tile-image-3': '"../static/homepage-images/tile-3-small.jpg"',
            'tile-image-4': '"../static/homepage-images/tile-4-small.jpg"',
            'tile-image-5': '"../static/homepage-images/tile-5-small.jpg"',
            'tile-image-6': 'null'
        }
        options.update(user_colors)

        return options

    @property
    def post_imports(self):
        return [
            'town'
        ]

    @property
    def extra_search_paths(self):
        return [module_path('onegov.town.theme', 'styles')]



""" Renders a onegov.page. """

import morepath

from onegov.core.security import Public, Private
from onegov.page import Page, PageCollection
from onegov.town import _
from onegov.town.app import TownApp
from onegov.town.elements import Link
from onegov.town.layout import PageLayout
from onegov.town.model import LinkEditor, PageEditor


@TownApp.html(model=Page, template='page.pt', permission=Public)
def view_page(self, request):

    if not request.is_logged_in:
        return view_public_page(self, request)
    else:
        return view_private_page(self, request)


def view_public_page(self, request):
    if self.meta['type'] == 'link':
        return morepath.redirect(self.content['url'])
    else:
        return {
            'layout': PageLayout(self, request),
            'title': self.title,
            'page': self,
            'children': [
                Link(child.title, request.link(child))
                for child in self.children
            ]
        }


def view_private_page(self, request):
    return {
        'layout': PageLayout(self, request),
        'title': self.title,
        'page': self,
        'add_links': tuple(get_links(self, request, 'add')),
        'edit_links': tuple(get_links(self, request, 'edit')),
        'children': [
            Link(child.title, request.link(child)) for child in self.children
        ]
    }


@TownApp.view(model=Page, request_method='DELETE', permission=Private)
def delete_page(self, request):
    request.assert_valid_csrf_token()

    if self.meta['type'] == 'page':
        message = _(u"The page was deleted")
    else:
        message = _(u"The link was deleted")

    PageCollection(request.app.session()).delete(self)
    request.success(message)


def get_links(self, request, action):

    if self.meta['type'] in ('town-root', 'page') and action == 'add':
        yield Link(
            _("Page"), request.link(PageEditor('new', self)),
            classes=('new-page', )
        )

        yield Link(
            _("Link"), request.link(LinkEditor('new', self)),
            classes=('new-link', )
        )

    if self.meta['type'] == 'town-root' and action == 'edit':
        yield Link(
            _("Edit"), request.link(PageEditor('edit', self)),
            classes=('edit-page', )
        )

    if self.meta['type'] == 'page' and action == 'edit':
        yield Link(
            _("Edit"), request.link(PageEditor('edit', self)),
            classes=('edit-page', )
        )

        if self.children:
            extra_warning = _(
                "Please note that this page has subpages "
                "which will also be deleted!"
            )
        else:
            extra_warning = ""

        yield Link(
            _("Delete"), request.link(self), request_method='DELETE',
            classes=('confirm', 'delete-page'),
            attributes={
                'data-confirm': _(
                    "Do you really want to delete the page \"${title}\"?",
                    mapping={
                        'title': self.title
                    }
                ),
                'data-confirm-yes': _("Delete Page"),
                'data-confirm-no': _("Cancel"),
                'data-confirm-extra': extra_warning,
                'redirect-after': request.link(self.parent)
            },
        )

    if self.meta['type'] == 'link' and action == 'edit':
        yield Link(
            _("Edit"), request.link(LinkEditor('edit', self)),
            classes=('edit-link', )
        )
        yield Link(
            _("Delete"), request.link(self), request_method='DELETE',
            classes=('confirm', 'delete-link'),
            attributes={
                'data-confirm': _(
                    "Do you really want to delete the link \"${title}\"?",
                    mapping={
                        'title': self.title
                    }
                ),
                'data-confirm-yes': _("Delete Link"),
                'data-confirm-no': _("Cancel"),
                'redirect-after': request.link(self.parent)
            }
        )



""" Contains a list of manual translations. Those are usually messages that
exist in external passages which are not queries for i18n extraction.

"""

from onegov.town import _

# messages defined in wtforms-components
_(u"Not a valid color.")



from cached_property import cached_property
from onegov.core.compat import zip_longest
from onegov.core.static import StaticFile
from onegov.page import Page, PageCollection
from onegov.town import _
from onegov.town.elements import Link
from onegov.town.model import ImageCollection
from purl import URL


class Layout(object):
    """ Contains methods to render a page inheriting from layout.pt.

    All pages inheriting from layout.pt rely on this class being present
    as 'layout' variable::

     @TownApp.html(model=Example, template='example.pt', permission=Public)
        def view_example(self, request):
            return { 'layout': DefaultLayout(self, request) }

    It is meant to be extended for different parts of the site. For example,
    the :class:`DefaultLayout` includes the top navigation defined by
    onegov.page.

    It's possible though to have a different part of the website use a
    completely different top navigation. For that, a new Layout class
    inheriting from this one should be added.

    """

    def __init__(self, model, request):
        self.model = model
        self.request = request

        # always include the common js files
        self.request.include('common')

    @property
    def town(self):
        """ An alias for self.request.app.town. """
        return self.request.app.town

    @cached_property
    def font_awesome_path(self):
        static_file = StaticFile.from_application(
            self.app, 'font-awesome/css/font-awesome.min.css')

        return self.request.link(static_file)

    @cached_property
    def app(self):
        """ Returns the application behind the request. """
        return self.request.app

    @cached_property
    def page_id(self):
        """ Returns the unique page id of the rendered page. Used to have
        a useful id in the body element for CSS/JS.

        """
        page_id = self.request.path_info
        page_id = page_id.lstrip('/')
        page_id = page_id.replace('/', '-')
        page_id = page_id.rstrip('-')

        return page_id or 'root'

    @cached_property
    def template_loader(self):
        """ Returns the chameleon template loader. """
        return self.app.registry._template_loaders['.pt']

    @cached_property
    def base(self):
        """ Returns the layout, which defines the base layout of all town
        pages. See ``templates/layout.pt``.

        """
        return self.template_loader['layout.pt']

    @cached_property
    def macros(self):
        """ Returns the macros, which offer often used html constructs.
        See ``templates/macros.pt``.

        """
        return self.template_loader['macros.pt']

    @cached_property
    def top_navigation(self):
        """ Returns a list of :class:`onegov.town.elements.Link` objects.
        Those links are used for the top navigation.

        If nothing is returned, no top navigation is displayed.

        """
        return None

    @cached_property
    def breadcrumbs(self):
        """ Returns a list of :class:`onegov.town.elements.Link` objects.
        Those links are used for the breadcrumbs.

        If nothing is returned, no top breadcrumbs are displayed.

        """
        return None

    @cached_property
    def sidebar_links(self):
        """ A list of links shown in the sidebar, used for navigation. """
        return None

    @cached_property
    def bottom_links(self):
        """ A list of links shown at the absolute bottom. Use this for
        links like administration, statistics, source-code.

        """
        return None

    def chunks(self, iterable, n, fillvalue=None):
        """ Iterates through an iterable, returning chunks with the given size.

        For example::

            chunks('ABCDEFG', 3, 'x') --> ABC DEF Gxx

        """

        args = [iter(iterable)] * n
        return zip_longest(fillvalue=fillvalue, *args)

    def csrf_protected_url(self, url):
        """ Adds a csrf token to the given url. """
        return URL(url).query_param('csrf-token', self.csrf_token).as_string()

    @cached_property
    def image_upload_url(self):
        """ Returns the url to the image upload action. """
        url = self.request.link(ImageCollection(self.app), name='upload')
        return self.csrf_protected_url(url)

    @cached_property
    def image_upload_json_url(self):
        """ Adds the json url for image uploads. """
        url = self.request.link(ImageCollection(self.app), name='upload.json')
        return self.csrf_protected_url(url)

    @cached_property
    def image_list_url(self):
        """ Adds the json url for image lists. """
        return self.request.link(ImageCollection(self.app), name='json')

    @cached_property
    def homepage_url(self):
        """ Returns the url to the main page. """
        return self.request.link(self.app.town)

    @cached_property
    def csrf_token(self):
        """ Returns a csrf token for use with DELETE links (forms do their
        own thing automatically).

        """
        return self.request.new_csrf_token()


class DefaultLayout(Layout):
    """ The defaut layout meant for the public facing parts of the site. """

    @cached_property
    def breadcrumbs(self):
        """ Returns the breadcrumbs for the current page. """
        return [Link(_("Homepage"), self.homepage_url)]

    @cached_property
    def root_pages(self):
        roots = PageCollection(self.app.session()).query().filter(
            Page.parent_id == None,
            Page.meta != None
        )
        return [r for r in roots.all() if r.meta.get('type') == 'town-root']

    @cached_property
    def top_navigation(self):
        return tuple(
            Link(r.title, self.request.link(r)) for r in self.root_pages
        )

    @cached_property
    def bottom_links(self):

        request = self.request

        if request.current_role == 'editor':
            return [
                Link(_(u'Logout'), request.link(self.town, 'logout')),
                Link(_(u'Images'), request.link(ImageCollection(self.app))),
                Link(u'OneGov', 'http://www.onegov.ch'),
                Link(u'Seantis GmbH', 'https://www.seantis.ch')
            ]
        elif request.current_role == 'admin':
            return [
                Link(_(u'Logout'), request.link(self.town, 'logout')),
                Link(_(u'Images'), request.link(ImageCollection(self.app))),
                Link(_(u'Settings'), request.link(self.town, 'einstellungen')),
                Link(u'OneGov', 'http://www.onegov.ch'),
                Link(u'Seantis GmbH', 'https://www.seantis.ch')
            ]
        else:
            return [
                Link(_(u'Login'), request.link(self.town, 'login')),
                Link(u'OneGov', 'http://www.onegov.ch'),
                Link(u'Seantis GmbH', 'https://www.seantis.ch')
            ]


class PageLayout(DefaultLayout):
    """ The default layout, extended with the navigation of page objects. """

    @cached_property
    def breadcrumbs(self):
        """ Returns the breadcrumbs for the current page. """

        return self.get_page_breadcrumbs(self.model)

    def get_page_breadcrumbs(self, page):
        links = [Link(_("Homepage"), self.homepage_url)]

        for ancestor in page.ancestors:
            links.append(Link(ancestor.title, self.request.link(ancestor)))

        links.append(Link(page.title, self.request.link(page)))

        return links

    @cached_property
    def sidebar_links(self):
        """ Returns the sidebar links for the current page. """

        links = []

        for page in self.model.siblings.all():
            if page != self.model:
                links.append(
                    Link(page.title, self.request.link(page))
                )
            else:
                links.append(
                    Link(page.title, self.request.link(page), active=True)
                )

                for page in self.model.children:
                    links.append(
                        Link(
                            page.title,
                            self.request.link(page),
                            classes=('childpage', )
                        )
                    )

                links.append(
                    Link("...", '#', classes=('new-content-placeholder',))
                )

        return links


class EditorLayout(PageLayout):

    def __init__(self, model, request, site_title):
        super(PageLayout, self).__init__(model, request)
        self.site_title = site_title

        self.request.include('redactor')
        self.request.include('redactor_theme')
        self.request.include('editor')

    @cached_property
    def sidebar_links(self):
        pass

    @cached_property
    def breadcrumbs(self):
        links = self.get_page_breadcrumbs(self.model.page)
        links.append(Link(self.site_title, url='#'))
        return links



""" Implements the adding/editing/removing of pages. """

import morepath

from onegov.core.security import Private
from onegov.form import Form, with_options
from onegov.page import PageCollection
from onegov.town import _
from onegov.town.utils import sanitize_html
from onegov.town.app import TownApp
from onegov.town.layout import EditorLayout
from onegov.town.model import LinkEditor, PageEditor
from wtforms import StringField, TextAreaField, validators
from wtforms.fields.html5 import URLField
from wtforms.widgets import TextArea


class BaseForm(Form):
    title = StringField(_("Title"), [validators.InputRequired()])


class LinkForm(BaseForm):
    url = URLField(_("URL"), [validators.InputRequired()])


class PageForm(BaseForm):
    lead = TextAreaField(
        label=_("Lead"),
        description=_("Describes what this page is about"),
        widget=with_options(TextArea, rows=4))

    text = TextAreaField(
        label=_(u"Text"),
        widget=with_options(TextArea, class_='editor'),
        filters=[sanitize_html])


@TownApp.form(
    model=PageEditor, form=PageForm, template='form.pt', permission=Private
)
def handle_page_form(self, request, form):

    if self.action == 'new':
        return handle_new_page(self, request, form, page_type='page')
    elif self.action == 'edit':
        return handle_edit_page(self, request, form, page_type='page')
    else:
        raise NotImplementedError


@TownApp.form(
    model=LinkEditor, form=LinkForm, template='form.pt', permission=Private
)
def handle_link_form(self, request, form):

    if self.action == 'new':
        return handle_new_page(self, request, form, page_type='link')
    elif self.action == 'edit':
        return handle_edit_page(self, request, form, page_type='link')
    else:
        raise NotImplementedError


def handle_new_page(self, request, form, page_type):
    assert page_type in {'page', 'link'}

    if page_type == 'page':
        site_title = _(u"New Page")
    else:
        site_title = _(u"New Link")

    if form.submitted(request):
        pages = PageCollection(request.app.session())

        if page_type == 'page':
            meta = {'type': 'page'}
            content = {
                'lead': form.lead.data,
                'text': form.text.data
            }
            message = _(u"Added a new page.")
        else:
            meta = {'type': 'link'}
            content = {'url': form.url.data}
            message = _(u"Added a new link.")

        page = pages.add(
            parent=self.page,
            title=form.title.data,
            meta=meta,
            content=content
        )

        request.success(message)

        return morepath.redirect(request.link(page))

    return {
        'layout': EditorLayout(self, request, site_title),
        'title': site_title,
        'form': form
    }


def handle_edit_page(self, request, form, page_type):
    assert page_type in {'page', 'link'}

    if form.submitted(request):
        self.page.title = form.title.data

        if page_type in {'page', 'town-root'}:
            self.page.content['lead'] = form.lead.data
            self.page.content['text'] = form.text.data
        else:
            self.page.content['url'] = form.url.data

        request.success(_(u"Your changes were saved."))

        return morepath.redirect(request.link(self.page))
    else:
        form.title.data = self.page.title

        if page_type == 'page':
            site_title = _(u"Edit Page")
            form.lead.data = self.page.content.get('lead', '')
            form.text.data = self.page.content.get('text', '')
        else:
            site_title = _(u"Edit Link")
            form.url.data = self.page.content.get('url', '')

    return {
        'layout': EditorLayout(self, request, site_title),
        'title': site_title,
        'form': form
    }



import bleach


def sanitize_html(html):
    """ Takes the given html and strips all but a whitelisted number of tags
    from it.

    """

    if not html:
        return html

    allowed_tags = [
        'a',
        'abbr',
        'b',
        'br',
        'blockquote',
        'code',
        'del',
        'div',
        'em',
        'i',
        'img',
        'hr',
        'li',
        'ol',
        'p',
        'strong',
        'sup',
        'ul',
        'h1',
        'h2',
        'h3',
        'h4',
        'h5',
        'h6'
    ]

    allowed_attributes = {
        'a': ['href', 'title'],
        'abbr': ['title'],
        'acronym': ['title'],
        'img': ['src', 'alt', 'title']
    }

    return bleach.clean(html, tags=allowed_tags, attributes=allowed_attributes)



import more.webassets
import onegov.core
import onegov.town
import os.path
import pytest
import transaction

from morepath import setup
from onegov.town.initial_content import add_initial_content
from onegov.user import UserCollection
from uuid import uuid4


@pytest.yield_fixture(scope="function")
def town_app(postgres_dsn, temporary_directory):
    config = setup()
    config.scan(more.webassets)
    config.scan(onegov.core)
    config.scan(onegov.town)
    config.commit()

    app = onegov.town.TownApp()
    app.namespace = 'test_' + uuid4().hex
    app.configure_application(
        dsn=postgres_dsn,
        filestorage='fs.osfs.OSFS',
        filestorage_options={
            'root_path': os.path.join(temporary_directory, 'file-storage'),
            'create': True
        },
        identity_secure=False,
        disable_memcached=True
    )
    app.set_application_id(app.namespace + '/' + 'test')

    add_initial_content(app.session(), 'Govikon')

    users = UserCollection(app.session())
    users.add('admin@example.org', 'hunter2', 'admin')
    users.add('editor@example.org', 'hunter2', 'editor')

    transaction.commit()

    yield app



""" Contains the paths to the different models served by onegov.town. """

from onegov.town.app import TownApp
from onegov.town.model import (
    Image,
    ImageCollection,
    Thumbnail,
    LinkEditor,
    PageEditor,
    Town
)
from onegov.page import Page, PageCollection


@TownApp.path(model=Town, path='/')
def get_town(app):
    return app.town


@TownApp.path(model=Page, path='/themen', absorb=True)
def get_page(app, absorb):
    return PageCollection(app.session()).by_path(absorb)


@TownApp.path(model=ImageCollection, path='/bilder')
def get_images(app):
    return ImageCollection(app)


@TownApp.path(model=Image, path='/bild/{filename}')
def get_image(app, filename):
    return ImageCollection(app).get_image_by_filename(filename)


@TownApp.path(model=Thumbnail, path='/thumbnails/{filename}')
def get_thumbnail(app, filename):
    return ImageCollection(app).get_thumbnail_by_filename(filename)


@TownApp.path(model=PageEditor, path='/editor/page/{page_id}/{action}')
def get_page_editor(app, page_id, action):
    page = PageCollection(app.session()).by_id(page_id)

    if page is None:
        return

    return PageEditor(action=action, page=page)


@TownApp.path(model=LinkEditor, path='/editor/link/{page_id}/{action}')
def get_link_editor(app, page_id, action):
    page = PageCollection(app.session()).by_id(page_id)

    if page is None:
        return

    return LinkEditor(action=action, page=page)



""" The settings view, defining things like the logo or color of the town. """

from onegov.core.security import Secret
from onegov.form import Form, with_options
from wtforms import StringField, TextAreaField, validators
from wtforms.widgets import TextArea
from wtforms_components import ColorField
from onegov.town import _
from onegov.town.app import TownApp
from onegov.town.elements import Link
from onegov.town.layout import DefaultLayout
from onegov.town.model import Town


class SettingsForm(Form):
    """ Defines the settings form for onegov town. """

    name = StringField(_("Name"), [validators.InputRequired()])
    logo_url = StringField(
        label=_("Logo"),
        description=_("URL pointing to the logo")
    )
    primary_color = ColorField(_("Primary Color"))
    homepage_images = TextAreaField(
        label=_("Homepage Images"),
        description=_(
            "Up to six URLs pointing to images for the tiles on the homepage."
        ),
        widget=with_options(TextArea, rows=6)
    )

    @property
    def theme_options(self):
        options = {
            'primary-color': self.primary_color.data.get_hex(),
        }

        urls = (url.strip() for url in self.homepage_images.data.split('\n'))
        urls = (url for url in urls if url)

        for ix, url in enumerate(urls):
            # put the url in apostrophes, because it's a sass string
            options['tile-image-{}'.format(ix + 1)] = '"{}"'.format(url)

        return options

    @theme_options.setter
    def theme_options(self, theme_options):
        self.primary_color.data = theme_options.get('primary-color')

        tile_image_keys = ('tile-image-{}'.format(ix) for ix in range(1, 7))
        urls = (theme_options.get(key) for key in tile_image_keys)
        urls = (url.strip('"') for url in urls if url)

        self.homepage_images.data = '\n'.join(urls)


@TownApp.form(
    model=Town, name='einstellungen', template='form.pt', permission=Secret,
    form=SettingsForm)
def handle_settings(self, request, form):
    """ Handles the GET and POST login requests. """

    if form.submitted(request):
        with request.app.update_town() as town:
            town.name = form.name.data
            town.logo_url = form.logo_url.data
            town.theme_options = form.theme_options

        request.success(_(u"Your changes were saved."))
    else:
        form.name.data = self.name
        form.logo_url.data = self.logo_url
        form.theme_options = self.theme_options

    layout = DefaultLayout(self, request)
    layout.breadcrumbs = [
        Link(_("Homepage"), layout.homepage_url),
        Link(_("Settings"), request.link(self))
    ]

    return {
        'layout': layout,
        'title': _(u'Settings'),
        'form': form
    }



""" The onegov town homepage. """

from collections import namedtuple
from onegov.core.security import Public
from onegov.town.app import TownApp
from onegov.town.elements import Link
from onegov.town.model import Town
from onegov.town.layout import DefaultLayout


@TownApp.html(model=Town, template='homepage.pt', permission=Public)
def view_town(self, request):
    """ Renders the town's homepage. """

    layout = DefaultLayout(self, request)
    Tile = namedtuple('Tile', ['page', 'links', 'number'])
    tiles = [
        Tile(
            page=Link(page.title, request.link(page)),
            number=ix + 1,
            links=[
                Link(c.title, request.link(c), classes=('tile-sub-link',))
                for c in page.children[:3]
            ]
        ) for ix, page in enumerate(layout.root_pages)
    ]

    return {
        'layout': layout,
        'title': self.name,
        'tiles': tiles
    }



""" Contains the models directly provided by onegov.town. """

import PIL

from onegov.core.filestorage import FilestorageFile
from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import JSON, UUID
from onegov.town.theme import user_colors
from sqlalchemy import Column, Text
from uuid import uuid4
from webob.exc import HTTPUnsupportedMediaType


class Town(Base, TimestampMixin):
    """ Defines the basic information associated with a town.

    It is assumed that there's only one town record in the schema associated
    with this town.
    """

    __tablename__ = 'towns'

    #: the id of the town, an automatically generated uuid
    id = Column(UUID, nullable=False, primary_key=True, default=uuid4)

    #: the name of the town (as registered with the Swiss governement)
    name = Column(Text, nullable=False)

    #: the logo of the town
    logo_url = Column(Text, nullable=True)

    #: the theme options of the town
    theme_options = Column(JSON, nullable=True, default=user_colors.copy)


class ImageCollection(object):
    """ Defines the collection of images uploaded to the site. Currently
    this is done without any ORM backing (and therefore without any
    special features like tagging, metadata and so on).

    Instead it's simply a list of images in a directory.

    This can be made more powerful (and complicated) once we have sufficent
    time to do it.

    """

    thumbnail_dimension = (256, 256)
    thumbnail_quality = 80
    allowed_extensions = {'png', 'jpg', 'jpeg', 'gif', 'svg'}
    thumbnail_extensions = {'png', 'jpg', 'jpeg', 'gif'}

    def __init__(self, app):
        assert app.has_filestorage

        self.path_prefix = 'images/'
        self.image_storage = app.filestorage.makeopendir('images')
        self.thumbnail_storage = self.image_storage.makeopendir('thumbnails')

    @property
    def images(self):
        """ Returns the :class:`~onegov.town.model.Image` instances in this
        collection.

        """
        images = self.image_storage.ilistdirinfo(files_only=True)
        images = sorted(images, key=lambda i: i[1]['created_time'])

        for filename, info in images:
            yield Image(filename, info)

    @property
    def thumbnails(self):
        """ Returns the :class:`~onegov.town.model.Thumbnail` instances in this
        collection.

        """
        images = self.thumbnail_storage.ilistdirinfo(files_only=True)
        images = sorted(images, key=lambda i: i[1]['created_time'])

        for filename, info in images:
            yield Thumbnail(filename, info)

    def store_image(self, image, filename):
        """ Stores an image (a file with a ``read()`` method) with the given
        filename. Note that these images are public, so the filename *should*
        be random.

        See :func:`onegov.core.filestorage.random_filename`.

        """
        extension = filename.split('.')[-1]

        if extension not in self.allowed_extensions:
            raise HTTPUnsupportedMediaType()

        self.image_storage.setcontents(filename, image.read())

        return self.get_image_by_filename(filename)

    def get_image_by_filename(self, filename):
        """ Returns the :class:`~onegov.town.model.Image` instance with the
        given name, or None if not found.

        """
        if self.image_storage.exists(filename):
            return Image(filename)

    def get_thumbnail_by_filename(self, filename):
        """ Returns the :class:`~onegov.town.model.Thumbnail` instance with the
        given name, or None if not found.

        This method *generates* the thumbnail if it doesn't exist yet!

        """
        if not self.image_storage.exists(filename):
            return None

        if self.thumbnail_storage.exists(filename):
            return Thumbnail(filename)

        extension = filename.split('.')[-1]

        if extension not in self.thumbnail_extensions:
            return Image(filename)

        im = PIL.Image.open(self.image_storage.getsyspath(filename))
        im.thumbnail(self.thumbnail_dimension)
        im.save(
            self.thumbnail_storage.open(filename, 'wb'),
            PIL.Image.EXTENSION['.' + extension],
            quality=self.thumbnail_quality
        )

        return Thumbnail(filename)

    def delete_image_by_filename(self, filename):
        """ Deletes both the image and the thumbnail of the given filename. """

        if self.image_storage.exists(filename):
            self.image_storage.remove(filename)

        if self.thumbnail_storage.exists(filename):
            self.thumbnail_storage.remove(filename)


class ImageFile(FilestorageFile):
    """ A filestorage file that points to an image. """

    def __init__(self, filename, info=None):
        self.filename = filename
        self.info = info or {}

    @property
    def date(self):
        if 'date' in self.info:
            return self.info['created_time'].date()


class Image(ImageFile):
    """ A filestorage file that points to a full image (not a thumbnail). """

    @property
    def thumbnail(self):
        return Thumbnail(self.filename)

    @property
    def path(self):
        return 'images/' + self.filename


class Thumbnail(ImageFile):
    """ A filestorage file that points to a thumbnail or the original file
    storage file, if there can't be a thumbnail (say for *.svg).

    Thumbnails are created on the fly and cached.

    """

    @property
    def image(self):
        return Image(self.filename)

    @property
    def path(self):
        return 'images/thumbnails/' + self.filename


class Editor(object):
    """ Defines the model for the page editor. Required because pages need
    to be edited outside their url structure, since their urls are absorbed
    completely and turned into SQL queries.

    """
    def __init__(self, action, page):
        """ The editor is defined by an action and a page/context.

        :action:
            One of 'new', 'edit' or 'delete'.

        :page:
            The 'context' of the action. The actual page in the case of 'edit'
            and 'delete'. The parent in the case of 'new'.

        """

        self.action = action
        self.page = page

    @property
    def page_id(self):
        return self.page.id


class PageEditor(Editor):
    """ Editor for the page type "page". """
    pass


class LinkEditor(Editor):
    """ Editor for the page type "link". """
    pass



from onegov.page import PageCollection
from onegov.town.model import Town


def add_initial_content(session, town_name):
    """ Adds the initial content for the given town on the given session.

    All content that comes with a new town is added here.

    """

    # can only be called if no town is defined yet
    assert not session.query(Town).first()

    session.add(Town(name=town_name))

    pages = PageCollection(session)

    pages.add_root(
        'Leben & Wohnen',
        name='leben-wohnen',
        meta={'type': 'town-root'}
    ),
    pages.add_root(
        'Kultur & Freizeit',
        name='kultur-freizeit',
        meta={'type': 'town-root'}
    ),
    pages.add_root(
        'Bildung & Gesellschaft',
        name='bildung-gesellschaft',
        meta={'type': 'town-root'}
    ),
    pages.add_root(
        'Gewerbe & Tourismus',
        name='gewerbe-tourismus',
        meta={'type': 'town-root'}
    ),
    pages.add_root(
        'Politik & Verwaltung',
        name='politik-verwaltung',
        meta={'type': 'town-root'}
    )

    session.flush()
